
// includes
#include "TGAImage.hpp"

using namespace std;

namespace bleifrei {
	namespace img {

		#pragma pack(push, 1)

		// Truevision Targa
		struct TGAColorMapInfo
		{
			unsigned short origin;
			unsigned short length;
			unsigned char bpp;
		};

		struct TGAHeader
		{
			unsigned char idlen;
			unsigned char colType;
			unsigned char imType;
			TGAColorMapInfo colormapinfo;
			unsigned short x;
			unsigned short y;
			unsigned short width;
			unsigned short height;
			unsigned char bpp;
			unsigned char imdesc;
		};

		#pragma pack(pop)

		/** Load a TGA file from a stream.
		 * Interpret data in stream as a Targa file and use it to create a new Image object.
		 * @param "istream &is" Source stream
		 * @throws StreamException
		 * @throws FileFormatException
		 * @throws OutOfMemoryException
		 */
		TGAImage::TGAImage(istream &is)
		{
			TGAHeader tgahead;

			// Read TGA header
			is.read((char *)&tgahead, sizeof(tgahead));

			// Throw exception if error occurred
			if (!is)
			{
				throw stream_exception();
			}

			// Initialize image
			create(tgahead.width, tgahead.height, bitDepthToImageFormatRev(tgahead.bpp));

			// Ignore identification field
			if (tgahead.idlen)
			{
				is.ignore(tgahead.idlen);
			}

			// Does the image have a palette?
			if (tgahead.colType)
			{
				int palbytes = ((tgahead.colormapinfo.length - tgahead.colormapinfo.origin) * tgahead.colormapinfo.bpp + 7) / 8;
				is.ignore(palbytes);
			}

			if (!is)
			{
				throw stream_exception();
			}

			// Decode image data
			switch (tgahead.imType)
			{
			// Uncompressed
			case 1:
			case 2:
			case 3:
				is.read((char *)m_imageData, m_bytesPerLine * getHeight());
				break;

			// RLE encoded
			case 9:
			case 10:
			case 11:
				decodeRLE(is, m_bytesPerLine * getHeight());
				break;

			// Unknown file format
			default:
				throw file_format_exception();
			}
/*
			// Convert image to RGB format
			switch (m_format) {
			case ImageFormatR8G8B8A8:
				convertBGRtoRGB32();
				break;
			case ImageFormatR8G8B8:
				convertBGRtoRGB24();
				break;
			default:
				convertBGRtoRGB();
			}
*/			
			if (!is) {
				throw stream_exception();
			}

			//flip(); // The eternal question... to flip, or not to flip TGAs
		}

		void TGAImage::save(ostream &os, ImageCompression compression)
		{
			TGAHeader tgahead;

			switch (m_format) {
			case ImageFormatR8G8B8A8 :
				convertBGRtoRGB32();
				break;
			case ImageFormatR8G8B8 :
				convertBGRtoRGB24();
				break;
			case ImageFormatR5G5B5A1 :
			case ImageFormatR5G6B5 :
				convertBGRtoRGB();
			}

			tgahead.width = m_width;
			tgahead.height = m_height;
			tgahead.bpp = imageFormatToBitDepth(m_format);
			tgahead.idlen = 0;
			tgahead.colType = 0;
			tgahead.imType = compression == ImageCompressionRaw ? 2 : 10;

			os.write((char *)&tgahead, sizeof(tgahead));

			if (compression == ImageCompressionRaw) {
				os.write((char *)getData(), this->getBytesPerLine() * getHeight());
			}
			else {
				encodeRLE(os);
			}
		}

	}
}